/**
 * BH1750_LightSensor
 * Klasse fuer den Zugriff auf einen Lichtsensor mit I2C.
 *
 * @mc       Arduino/ESP8266
 * @autor    Harald Sattler
 * @version  1.0
 * @created  17.01.2019
 *
 */

#include "BH1750_LightSensor.h"
#include <Wire.h> //BH1750 IIC Mode 

//#define DEBUG
#include "Debug.h"

/**
 * Initialisierung mit der I2C-Adresse, unter der der Sensor erreichbar ist.
 * Die Maximalwerte vom Sensor koennen automatisch
 * ausgemessen werden (wir verwenden denselben define LDR_AUTOSCALE).
 * Ansonsten muss man diese Werte im #define-DEBUG-Mode
 * ausmessen und eintragen.
 */
BH1750::BH1750(byte adr, boolean isInverted, uint8_t constrainMin, uint8_t constrainMax, boolean autoScale, uint16_t ldrHysterese, uint16_t minLdr, uint16_t maxLdr) {
	_adr = adr;
	_isInverted = isInverted;
	_autoScale = autoScale;
	_ldrMin = minLdr;
	_ldrMax = maxLdr;
	_constrainMin = constrainMin;
	_constrainMax = constrainMax;
	_ldrHysteresis = ldrHysterese;
	_lastValue = 1;
	_outputValue = 0;
	if (_autoScale) {
		_min = 1023;
		_max = 0;
	} else {
		_min = _ldrMin;
		_max = _ldrMax;
	}
}

/**
 * Welchen Wert liefert der Sensor? In Prozent...
 */
byte BH1750::value() {
	int rawVal, val;
	long sensor_val = 0;
	//byte buff[2];
	int i = 0;
	if (_messungLaeuft == false) {
		_messungLaeuft = true;

		Wire.beginTransmission(_adr);
		//Wire.write(0x10);					// 1 lx resolution 120ms, continuous
		Wire.write(0x20);					// 1 lx resolution 120ms, single measure
		Wire.endTransmission();
		_measurementStarted = millis();		// Start-Zeitpunkt der Messung
	}
	if ((_messungLaeuft == true) && (millis() > _measurementStarted + 200)) {	// warte 180ms auf Messung-Ende
		_messungLaeuft = false;

		if (BH1750_Read(_adr) == 2) {
			// der Faktor 1.2 ist die Messgenauigkeit des Sensors. Der aus den Registern gelesene Wert muss
			// durch 1.2 geteilt werden, um die Helligkeit in Lux zu erhalten. (0.96..1.44, typisch 1.2)
			sensor_val = ((_buff[0] << 8) | _buff[1]) / 1.2;	// Rechenvorschrift lt. Datenblatt
			DEBUG_PRINT(F("sensor_val: ")); DEBUG_PRINT(sensor_val);
			_RAWvalue = sensor_val;			// deHarry, 2020-03-30, fuer debugging

			// bei Dunkelheit gibt der Sensor 0 lux aus, das ist zu wenig, deshalb wird die Untergrenze beim
			//	Mapping angehoben (200). Der Wert muss ggf. getuned werden
			// Hinter Microwood ist der Lichtpegel sehr eingeschrnkt, deshalb wird die Obergrenze (urspr. 10000)
			//	einstellbar (ber die GUI, Settings) gemacht, so dass der Regelbereich zur Laufzeit angepasst werden kann. ->>> nicht implementiert!!!!
			sensor_val = map((long)sensor_val, (long)0, (long)10000, (long)200, (long)1023);	// die lux-Werte des Sensors reichen von 0..65535

			sensor_val = constrain(sensor_val, 0, 1023);		// falls es mal heller werden sollte, begrenzen
			DEBUG_PRINT(F(" - map_sensor_val: ")); DEBUG_PRINTLN(sensor_val);

			if (!_isInverted) {
				rawVal = sensor_val;
			} else {
				rawVal = (1023 - sensor_val);
			}

			if ((rawVal != _lastValue)
				&& ((rawVal == 0) || (rawVal == 1023)
				|| (rawVal > (_lastValue + _ldrHysteresis) 
				|| (rawVal < _lastValue - _ldrHysteresis)))) {
				val = rawVal;
				_lastValue = val;
				if (_autoScale) {
					if (val < _min) {
						_min = val;
					}
					if (val > _max) {
						_max = val;
					}
				} else {
					val = constrain(val, _min, _max);
				}
				uint8_t mapVal = 50;
				if (_min != _max) {
					mapVal = map(val, _min, _max, 0, 255);
				}
				mapVal = constrain(mapVal, _constrainMin, _constrainMax);
				DEBUG_PRINT(F("rawVal: ")); DEBUG_PRINT(rawVal); 
				DEBUG_PRINT(F(" val: ")); DEBUG_PRINT(val); 
				DEBUG_PRINT(F(" _min: ")); DEBUG_PRINT(_min); 
				DEBUG_PRINT(F(" _max: ")); DEBUG_PRINT(_max); 
				DEBUG_PRINT(F(" mapValue: ")); DEBUG_PRINTLN(mapVal);
				_outputValue = mapVal;
			}
		}
	}
	return _outputValue;
}

long BH1750::lRAWvalue() {								// deHarry, 2020-03-30, fuer Anzeige bei debugVals (in der GUI)
	return _RAWvalue;
}

int BH1750::BH1750_Read(int address) //
{
	int i = 0;
	Wire.beginTransmission(address);
	Wire.requestFrom(_adr, (byte)2);	// 2 byte vom Sensor abrufen (der Sensor liefert 16bit Werte)
	while (Wire.available()) //
	{
		_buff[i] = Wire.read();  // receive one byte
		i++;
	}
	Wire.endTransmission();
	return i;
}
